#!/bin/sh
# Copyright 2023 Charadon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

if [ ! "$MSYSTEM" = "MSYS" ];
then
	echo "You must run MSYS2 with the MSYS runtime."
	exit 1
fi

help_screen() {
	echo "Usage: $0 [verb/file]"
	echo "Turns a POSIX shell script into a Windows Executable."
	printf "\n"
	echo "Valid Verbs:"
	printf "help                 Displays this screen.\n"
	printf "clean                Cleans the workspace.\n"
	printf "<path/to/script.sh>  Tells script which script to turn into executable.\n\n"
	echo "Exit status:"
	echo " 0 = Success."
	echo " 1 = Script failed before build part started."
	echo " 2 = Checksum mismatch of downloaded files."
	echo " 5 = Failed to change directory at some point in the script."
	printf "\n"
	echo "For more information visit sh2exe's website at: https://sh2exe.iotib.net"
	echo "or the git repo at: https://pagure.io/DaLazySlackers/sh2exe"
	exit "$1"
}

SEVEN_ZIP="7zr.exe"

if [ -z "$1" ];
then
	help_screen 1
elif [ "$1" = "help" ];
then
	help_screen 0
fi

set -eu

WORKDIR="$PWD/work"
rm -rf "$WORKDIR"
mkdir -p "$WORKDIR"
rm output.log > /dev/null 2>&1 || true
STARTING_DIR="$PWD"

COREUTILS_SHA512SUM="a6ee2c549140b189e8c1b35e119d4289ec27244ec0ed9da0ac55202f365a7e33778b1dc7c4e64d1669599ff81a8297fe4f5adbcc8a3a2f75c919a43cd4b9bdfa *coreutils-9.1.tar.xz"
COREUTILS_VERSION="9.1"
OKSH_SHA512SUM="1b412ebbae58bdb1ccf9b9e0b64bab233cd5e184a02bc67db9b4ba82fce0749b8166dd688c24fd890d22a111d7f6ba859c47e3175f28675fca9c6def8c416a08 *oksh-7.2.tar.gz"
OKSH_VERSION="7.2"
NANO_SHA512SUM="b1eb5c06460b1407e3fc959070d2eaaba29d236ef099654fcbf9ba7cf7b22f64ebc9e6f232d471601c35f0e1c78d8c7aad9269e2afed25c0899572b97d44a95c *nano-7.1.tar.gz"
NANO_VERSION="7.1"
LESS_SHA512SUM="7945b7f88921832ebb1b45fba8cbb449ee0133342796b654a52c146dfff3d84db18724ee84e53349eeea6017a0ebe2d8eb5366210275981dde7bb7190118fa66 *less-608.tar.gz"
LESS_VERSION="608"
GAWK_SHA512SUM="1d2aaf90bd074379e13dce65b53e80d449c383ff1f2862bb294644b2053383c19172d3b68bfd9caad82ccd11bea1baf627994280e845da9fdd0e773a000cfe40 *gawk-5.2.1.tar.gz"
GAWK_VERSION="5.2.1"
GREP_SHA512SUM="69cf162886770c4ff86f869f78d8a7e291480f576a92f2afa87c42290c7d34c4ca5cf52e47d10c8e310eadd605d209a2701dc714b651d122be9eca1e73130409 *grep-3.8.tar.gz"
GREP_VERSION="3.8"

# Speed up compilation times.
# Doesn't work: https://github.com/msys2/MSYS2-packages/issues/2492
#CC="ccache gcc"
#export CC
#CXX="ccache g++"
#export CXX



if [ "$1" = "clean" ];
then
	echo "Cleaning... "
	rm -rvf "$WORKDIR" || true
	rm -rvf "coreutils-$COREUTILS_VERSION/" || true
	rm -rvf "oksh-$OKSH_VERSION/" || true
	rm -rvf "nano-$NANO_VERSION/" || true
	rm -rvf "gawk-$GAWK_VERSION/" || true
	rm -rvf "grep-$GREP_VERSION/" || true
	rm -rvf "less-$LESS_VERSION/" || true
	rm -f ./*.tar.?z || true
	rm -f ./*.7z || true
	rm -f 7zip.h || true
	rm -f runtime.h || true
	rm -f 7zr.exe || true
	rm -f *.stackdump
	cd bin2c
	make clean
	cd ..
	exit 0
else
	SHEBANG="$(head -1 "$1")"
	if [ ! "$SHEBANG" = "#!/bin/sh" ];
	then
		echo "Shebang needs to be #!/bin/sh."
		exit 1
	fi
	SCRIPT_NAME="$1"
fi

echo "If your terminal closes, that means MSYS had an update that required a runtime restart."
echo "If the script ends, but there's no checkmark after the task, check the output.log for errors."
printf "Updating system... "
yes | pacman -Syu >> output.log 2>&1
printf "✓\n"
printf "Installing dependencies... "
yes | pacman -S ncurses-devel gcc texinfo ccache >> output.log 2>&1
printf "✓\n"

download_7zip() {
	curl "https://7-zip.org/a/7zr.exe" -O
}

build_coreutils() {
	curl "https://ftp.gnu.org/pub/gnu/coreutils/coreutils-$COREUTILS_VERSION.tar.xz" -O
	if [ ! "$(sha512sum coreutils-"$COREUTILS_VERSION".tar.xz)" = "$COREUTILS_SHA512SUM" ];
	then
		echo "Checksum mismatch for coreutils."
		exit 2
	fi
	if [ -f "coreutils-$COREUTILS_VERSION/DONE" ];
	then
		cd "coreutils-$COREUTILS_VERSION" || exit 5
		make DESTDIR="$WORKDIR" install-exec
		cd "$STARTING_DIR" || exit 5
		return 0
	fi
	tar -xvf "coreutils-$COREUTILS_VERSION".tar.xz
	cd "coreutils-$COREUTILS_VERSION" || exit 5
	make clean || true
	./configure --prefix=/usr
	make DESTDIR="$WORKDIR" -j "$(nproc)" install-exec
	touch DONE
	cd "$STARTING_DIR" || exit 5
	return 0
}

build_nano() {
	curl "https://ftp.gnu.org/gnu/nano/nano-$NANO_VERSION.tar.gz" -O
	if [ ! "$(sha512sum nano-$NANO_VERSION.tar.gz)" = "$NANO_SHA512SUM" ];
	then
		echo "Checksum mismatch for nano."
		exit 2
	fi
	if [ -f "nano-$NANO_VERSION/DONE" ];
	then
		cd "nano-$NANO_VERSION" || exit 5
		make DESTDIR="$WORKDIR" install-exec
		cd "$STARTING_DIR" || exit 5
		return 0
	fi
	tar -xvf "nano-$NANO_VERSION.tar.gz"
	cd "nano-$NANO_VERSION" || exit 5
	make clean || true
	./configure --prefix=/usr --enable-tiny
	make -j "$(nproc)"
	make DESTDIR="$WORKDIR" -j "$(nproc)" install
	touch DONE
	cd "$STARTING_DIR" || exit 5
	return 0
}

build_oksh() {
	curl "https://github.com/ibara/oksh/releases/download/oksh-$OKSH_VERSION/oksh-$OKSH_VERSION.tar.gz" -OL
	if [ ! "$(sha512sum oksh-$OKSH_VERSION.tar.gz)" = "$OKSH_SHA512SUM" ];
	then
		echo "Checksum mismatch for OKSH."
		exit 2
	fi
	if [ -f oksh-$OKSH_VERSION/DONE ];
	then
		cd "oksh-$OKSH_VERSION" || exit 5
		make DESTDIR="$WORKDIR" install
		cd "$STARTING_DIR" || exit 5
		return 0
	fi
	tar -xvf "oksh-$OKSH_VERSION.tar.gz"
	cd "oksh-$OKSH_VERSION" || exit 5
	make clean || true
	./configure --enable-curses --enable-sh --prefix=/usr
	make -j "$(nproc)"
	make DESTDIR="$WORKDIR" install
	touch DONE
	cd "$STARTING_DIR" || exit 5
	return 0
}

build_less() {
	curl "https://ftp.gnu.org/gnu/less/less-$LESS_VERSION.tar.gz" -O
	if [ ! "$(sha512sum less-$LESS_VERSION.tar.gz)" = "$LESS_SHA512SUM" ];
	then
		echo "Checksum mismatch for Less"
		exit 2
	fi
	if [ -f less-$LESS_VERSION/DONE ];
	then
		cd "less-$LESS_VERSION" || exit 5
		make DESTDIR="$WORKDIR" install
		cd "$STARTING_DIR" || exit 5
		return 0
	fi
	tar -xvf "less-$LESS_VERSION.tar.gz"
	cd "less-$LESS_VERSION" || exit 5
	make clean || true
	./configure --prefix=/usr
	make -j "$(nproc)"
	make DESTDIR="$WORKDIR" install
	touch DONE
	cd "$STARTING_DIR" || exit 5
	return 0
}

build_gawk() {
	curl "https://ftp.gnu.org/gnu/gawk/gawk-$GAWK_VERSION.tar.gz" -O
	if [ ! "$(sha512sum gawk-$GAWK_VERSION.tar.gz)" = "$GAWK_SHA512SUM" ];
	then
		echo "Checksum mismatch for Gawk"
		exit 2
	fi
	if [ -f gawk-$GAWK_VERSION/DONE ];
	then
		cd "gawk-$GAWK_VERSION" || exit 5
		make DESTDIR="$WORKDIR" install
		cd "$STARTING_DIR" || exit 5
		return 0
	fi
	tar -xvf "gawk-$GAWK_VERSION.tar.gz"
	cd "gawk-$GAWK_VERSION" || exit 5
	make clean || true
	./configure --prefix=/usr
	make -j "$(nproc)"
	make DESTDIR="$WORKDIR" install
	touch DONE
	cd "$STARTING_DIR" || exit 5
	return 0
}

build_grep() {
	curl "https://ftp.gnu.org/gnu/grep/grep-$GREP_VERSION.tar.gz" -O
	if [ ! "$(sha512sum grep-$GREP_VERSION.tar.gz)" = "$GREP_SHA512SUM" ];
	then
		echo "Checksum mismatch for Grep"
		exit 2
	fi
	if [ -f grep-$GREP_VERSION/DONE ];
	then
		cd "grep-$GREP_VERSION" || exit 5
		make DESTDIR="$WORKDIR" install
		cd "$STARTING_DIR" || exit 5
		return 0
	fi
	tar -xvf "grep-$GREP_VERSION.tar.gz"
	cd "grep-$GREP_VERSION" || exit 5
	make clean || true
	./configure --prefix=/usr
	make -j "$(nproc)"
	make DESTDIR="$WORKDIR" install
	touch DONE
	cd "$STARTING_DIR" || exit 5
	return 0
}

build_bin2c() {
	cd bin2c || exit 5
	make
	cd "$STARTING_DIR" || exit 5
}

cleanup_work() {
	# Remove man pages and headers.
	rm -rf "${WORKDIR:?}/usr/share/man"
	rm -rf "${WORKDIR:?}/usr/share/info"
	rm -rf "${WORKDIR:?}/usr/include"
	# Strip binaries
	for i in "$WORKDIR/usr/bin"/*.exe;
	do
		strip --strip-unneeded "$i"
	done
	# Add DLLs from MSYS so executables can run.
	cp "/usr/bin/msys-ncursesw6.dll" "$WORKDIR/usr/bin"
	cp "/usr/bin/msys-2.0.dll" "$WORKDIR/usr/bin"
	cp "/usr/bin/msys-iconv-2.dll" "$WORKDIR/usr/bin"
	cp "/usr/bin/msys-intl-8.dll" "$WORKDIR/usr/bin"
	cp "/usr/bin/msys-pcre-1.dll" "$WORKDIR/usr/bin"
	cp "/usr/bin/msys-gcc_s-seh-1.dll" "$WORKDIR/usr/bin"
	# Strip DLLs to save space.
	for i in "$WORKDIR/usr/bin"/*.dll;
	do
		strip --strip-unneeded "$i"
	done
	cp "$SCRIPT_NAME" "$WORKDIR/usr/bin"
	mkdir -p "$WORKDIR/usr/share/terminfo/63/"
	cp "/usr/share/terminfo/63/cygwin" "$WORKDIR/usr/share/terminfo/63/"
	# Time to 7zip it up.
	./$SEVEN_ZIP a -mx1 runtime.7z "$WORKDIR/*"
}

build_runner() {
	cp runsh.c "$WORKDIR/$SCRIPT_NAME.c"
	bin2c/bin2c.exe "$SEVEN_ZIP" "$WORKDIR/7zip.h" SevenZip
	bin2c/bin2c.exe "runtime.7z" "$WORKDIR/runtime.h" RuntimeArchive
	sed -i "s|INSERT_SCRIPT_NAME_HERE|$(basename "$SCRIPT_NAME")|g" "$WORKDIR/$SCRIPT_NAME.c"
	MSYSTEM="MINGW64" bash -l -c "gcc -I$WORKDIR -Wall -O1 \"$WORKDIR/$(basename $SCRIPT_NAME).c\" -o \"$WORKDIR/$(basename $SCRIPT_NAME).exe\""
}
printf "Downloading 7-zip... "
download_7zip >> output.log 2>&1
printf "✓\n"
printf "Compiling and installing GNU Coreutils... "
build_coreutils >> output.log 2>&1
printf "✓\n"
printf "Installing some of MSYS2's Coreutils to workaround bugs... "
install -Dm755 "/usr/bin/ls" "$WORKDIR/usr/bin/ls"
install -Dm755 "/usr/bin/du" "$WORKDIR/usr/bin/du"
printf "✓\n"
printf "Compiling and installing GNU Awk... "
build_gawk >> output.log 2>&1
printf "✓\n"
# Compiling grep and installing it results in segfault. We'll use MSYS2's for now.
printf "Installing GNU Grep... "
install -Dm755 "/usr/bin/grep" "$WORKDIR/usr/bin/grep"
printf "✓\n"
printf "Compiling and installing GNU Nano Text Editor... "
build_nano >> output.log 2>&1
printf "✓\n"
printf "Compiling and installing GNU Less... "
build_less >> output.log 2>&1
printf "✓\n"
printf "Compiling and installing Portable OpenBSD ksh... "
build_oksh >> output.log 2>&1
printf "✓\n"
printf "Compiling bin2c... "
build_bin2c >> output.log 2>&1
printf "✓\n"
printf "Creating runtime archive... "
cleanup_work >> output.log 2>&1
printf "✓\n"
printf "Compiling executable... "
build_runner >> output.log 2>&1
printf "✓\n"
echo "If all went well, your new executable should be in $WORKDIR/$(basename $SCRIPT_NAME).exe"
exit 0
